
curl https://www.antlr.org/download/antlr-4.8-complete.jar -o antlr.jar
curl https://www.antlr.org/download/antlr4-cpp-runtime-4.8-source.zip -o cpp_runtime.zip
unzip -o cpp_runtime.zip -d cpp_runtime